from __future__ import annotations

from restdoctor.utils.pydantic import convert_pydantic_errors_to_drf_errors


def test_convert_pydantic_errors_to_drf_errors():
    pydantic_errors = [
        {'loc': ('created_at',), 'msg': 'invalid datetime format', 'type': 'value_error.datetime'},
        {'loc': ('field_a',), 'msg': 'str type expected', 'type': 'type_error.str'},
        {
            'loc': ('field_b', 'inner_b'),
            'msg': 'value is not a valid integer',
            'type': 'type_error.integer',
        },
    ]
    expected_errors = {
        'created_at': 'invalid datetime format',
        'field_a': 'str type expected',
        'field_b.inner_b': 'value is not a valid integer',
    }
    drf_errors = convert_pydantic_errors_to_drf_errors(pydantic_errors)

    assert drf_errors == expected_errors



from __future__ import annotations

import pytest
from django.test import RequestFactory

from tests.test_unit.stubs import ModelAViewSet


@pytest.mark.django_db
def test_retrieve_get_item_success_case(mocker, resource_viewset_dispatch):
    resource_discriminator = 'one'
    mocker_get_item = mocker.patch.object(ModelAViewSet, 'get_item')
    view_func, mocked_dispatch = resource_viewset_dispatch(
        resource_discriminator, ModelAViewSet, actions={'get': 'retrieve'}
    )
    request = RequestFactory().get('/', {'view_type': resource_discriminator})

    view_func(request)

    assert mocker_get_item.called_once()
    assert mocked_dispatch.called_once()



from __future__ import annotations

import typing


def convert_pydantic_errors_to_drf_errors(
    pydantic_errors: list[dict[str, typing.Any]]
) -> dict[str, str]:
    drf_errors = {}
    for error in pydantic_errors:
        drf_errors['.'.join(error['loc'])] = error['msg']
    return drf_errors



from __future__ import annotations

import pytest

from restdoctor.utils.media_type import parse_accept, parse_accept_header


def test__parse_accept_header__none():
    result = parse_accept_header(None)

    assert result is None


@pytest.mark.parametrize(
    ('accept', 'expected_version'),
    [
        ('*/*', 'fallback'),
        ('text/html', 'fallback'),
        ('application/json', 'fallback'),
        ('application/vnd.vendor', 'default'),
    ],
)
def test__parse_accept__fallback(settings, accept, expected_version):
    settings.API_FALLBACK_VERSION = 'fallback'
    settings.API_DEFAULT_VERSION = 'default'
    settings.API_VERSIONS = {'v1': 'v1', 'v2': 'v2'}
    settings.API_FALLBACK_FOR_APPLICATION_JSON_ONLY = False

    result = parse_accept(accept)

    assert expected_version == result.version


@pytest.mark.parametrize(
    ('accept', 'expected_version'),
    [
        ('*/*', 'default'),
        ('text/html', 'default'),
        ('application/json', 'fallback'),
        ('application/vnd.vendor', 'default'),
    ],
)
def test__parse_accept__default(settings, accept, expected_version):
    settings.API_FALLBACK_VERSION = 'fallback'
    settings.API_DEFAULT_VERSION = 'default'
    settings.API_VERSIONS = {'v1': 'v1', 'v2': 'v2'}
    settings.API_FALLBACK_FOR_APPLICATION_JSON_ONLY = True

    result = parse_accept(accept)

    assert expected_version == result.version


@pytest.mark.parametrize(
    ('accept', 'expected_version', 'expected_discriminator'),
    [
        ('application/vnd.vendor', 'default', None),
        ('application/vnd.vendor.v1', 'v1', None),
        ('application/vnd.vendor.v1.full', 'v1', None),
        ('application/vnd.vendor.v1.full+json', 'v1', None),
        ('application/vnd.vendor.v1-extended', 'v1', 'extended'),
        ('application/vnd.vendor.v1-extended.full', 'v1', 'extended'),
        ('application/vnd.vendor.v1-extended.full+json', 'v1', 'extended'),
        ('application/vnd.vendor.v2', 'v2', None),
        ('application/vnd.vendor.v2.full', 'v2', None),
        ('application/vnd.vendor.v2.full+json', 'v2', None),
        ('application/vnd.vendor.v2-extended', 'v2', 'extended'),
        ('application/vnd.vendor.v2-extended.full', 'v2', 'extended'),
        ('application/vnd.vendor.v2-extended.full+json', 'v2', 'extended'),
        ('application/vnd.vendor.v3', 'default', None),
        ('application/vnd.vendor.v3.full', 'default', None),
        ('application/vnd.vendor.v3.full+json', 'default', None),
        ('application/vnd.vendor.v3-extended', 'default', 'extended'),
        ('application/vnd.vendor.v3-extended.full', 'default', 'extended'),
        ('application/vnd.vendor.v3-extended.full+json', 'default', 'extended'),
    ],
)
def test__parse_accept__versions_and_resource_discriminator(
    settings, accept, expected_version, expected_discriminator
):
    settings.API_FALLBACK_VERSION = 'fallback'
    settings.API_DEFAULT_VERSION = 'default'
    settings.API_RESOURCE_DEFAULT = 'common'
    settings.API_VERSIONS = {'v1': 'v1', 'v2': 'v2'}

    result = parse_accept(accept)

    assert expected_version == result.version
    assert expected_discriminator == result.resource_discriminator


@pytest.mark.parametrize(
    ('vendor', 'expected_vendor'), [(None, 'vendor'), ('', 'vendor'), ('my_vendor', 'my_vendor')]
)
def test__parse_accept__vendor_success_case(vendor, expected_vendor):
    result = parse_accept('application/vnd', vendor=vendor)

    assert expected_vendor == result.vendor



from __future__ import annotations

import copy
import datetime

import pytest
from pydantic import BaseModel, Field, StrictInt, StrictStr

from restdoctor.rest_framework.schema.generators import RefsSchemaGenerator
from restdoctor.rest_framework.schema.openapi import RestDoctorSchema
from restdoctor.rest_framework.schema.serializers import OPENAPI_REF_PREFIX
from restdoctor.rest_framework.serializers import PydanticSerializer


class PydanticTestModel(BaseModel):
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.utcnow)
    field_a: StrictStr
    field_b: StrictInt


class PydanticNestedTestModel(BaseModel):
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.utcnow)
    nested_field: PydanticTestModel


class TestPydanticSerializer(PydanticSerializer):
    pydantic_model = PydanticTestModel


class TestNestedPydanticSerializer(PydanticSerializer):
    pydantic_model = PydanticNestedTestModel


@pytest.fixture()
def test_model_schema():
    return {
        'description': 'PydanticTestModel',
        'type': 'object',
        'properties': {
            'created_at': {'description': 'Created At', 'type': 'string', 'format': 'date-time'},
            'field_a': {'description': 'Field A', 'type': 'string'},
            'field_b': {'description': 'Field B', 'type': 'integer'},
        },
        'required': ['field_a', 'field_b'],
    }


@pytest.fixture()
def test_nested_model_schema_without_definitions():
    return {
        'description': 'PydanticNestedTestModel',
        'type': 'object',
        'properties': {
            'created_at': {'description': 'Created At', 'type': 'string', 'format': 'date-time'},
            'nested_field': {'$ref': f'{OPENAPI_REF_PREFIX}PydanticTestModel'},
        },
        'required': ['nested_field'],
    }


@pytest.fixture()
def test_nested_model_schema(test_model_schema, test_nested_model_schema_without_definitions):
    schema = copy.deepcopy(test_nested_model_schema_without_definitions)
    schema['properties']['nested_field'] = {'$ref': '#/definitions/PydanticTestModel'}
    schema['definitions'] = {'PydanticTestModel': test_model_schema}
    return schema


def test_get_serializer_schema_success_case(test_model_schema):
    schema = RestDoctorSchema().get_serializer_schema(TestPydanticSerializer())

    assert schema == test_model_schema


def test_map_serializer_without_generator_success_case(test_model_schema):
    schema = RestDoctorSchema().map_serializer(TestPydanticSerializer())

    assert schema == test_model_schema


def test_map_serializer_with_generator_success_case(test_model_schema):
    schema_generator = RefsSchemaGenerator()
    schema = RestDoctorSchema(generator=schema_generator).map_serializer(TestPydanticSerializer())
    ref = f'{OPENAPI_REF_PREFIX}PydanticTestModel'

    assert schema == {'$ref': ref}
    assert schema_generator.local_refs_registry.get_local_ref(ref) == test_model_schema


def test_get_serializer_schema_for_nested_serializer_success_case(test_nested_model_schema):
    schema = RestDoctorSchema().get_serializer_schema(TestNestedPydanticSerializer())

    assert schema == test_nested_model_schema


def test_map_serializer_with_nested_serializer_success_case(test_nested_model_schema):
    schema = RestDoctorSchema().map_serializer(TestNestedPydanticSerializer())

    assert schema == test_nested_model_schema


def test_map_serializer_with_refs_generator_with_nested_serializer_success_case(
    test_nested_model_schema_without_definitions, test_model_schema
):
    schema_generator = RefsSchemaGenerator()
    schema = RestDoctorSchema(generator=schema_generator).map_serializer(
        TestNestedPydanticSerializer()
    )
    ref = f'{OPENAPI_REF_PREFIX}PydanticNestedTestModel'
    nested_ref = test_nested_model_schema_without_definitions['properties']['nested_field']['$ref']

    assert schema == {'$ref': ref}
    assert (
        schema_generator.local_refs_registry.get_local_ref(ref)
        == test_nested_model_schema_without_definitions
    )
    assert schema_generator.local_refs_registry.get_local_ref(nested_ref) == test_model_schema



from __future__ import annotations

import pytest
from django.core.validators import URLValidator
from django_filters import (
    BooleanFilter,
    ChoiceFilter,
    MultipleChoiceFilter,
    TypedChoiceFilter,
    TypedMultipleChoiceFilter,
    DateFilter,
    DateFromToRangeFilter,
    DateTimeFilter,
    DateTimeFromToRangeFilter,
    NumberFilter,
    TimeFilter,
)
from django_filters.rest_framework import BooleanFilter as RestBooleanFilter

from restdoctor.rest_framework.schema.filters import get_filter_schema, FILTER_MAP

url_pattern = URLValidator.regex.pattern.replace('\\Z', '\\z')


@pytest.mark.parametrize(
    ('field', 'expected_schema'),
    [
        (BooleanFilter(), {'type': 'boolean'}),
        (RestBooleanFilter(), {'type': 'boolean'}),
        (
            ChoiceFilter(choices=((1, 1), (2, 2)), empty_label=None),
            {'type': 'integer', 'enum': [1, 2]},
        ),
        (
            MultipleChoiceFilter(choices=((False, False), (True, True))),
            {'type': 'boolean', 'enum': [False, True]},
        ),
        (TypedChoiceFilter(choices=((1.1, 1.1), (2, 2))), {'type': 'number', 'enum': [1.1, 2]}),
        (
            TypedMultipleChoiceFilter(choices=(('a', 'a'), ('b', 'b'))),
            {'type': 'string', 'enum': ['a', 'b']},
        ),
        (DateFilter(), {'type': 'string', 'format': 'date'}),
        (DateFromToRangeFilter(), {'type': 'string', 'format': 'date'}),
        (DateTimeFilter(), {'type': 'string', 'format': 'date-time'}),
        (DateTimeFromToRangeFilter(), {'type': 'string', 'format': 'date-time'}),
        (NumberFilter(), {'type': 'number'}),
        (TimeFilter(), {'type': 'string', 'format': 'time'}),
    ],
)
def test__get_filter_schema(field, expected_schema):
    result = get_filter_schema(field)

    assert result == expected_schema


class CustomDateFilter(DateFilter):
    pass


def test__get_filter_schema__custom_field():
    result = get_filter_schema(
        filter_field=CustomDateFilter(),
        filter_map={**FILTER_MAP, CustomDateFilter: {'type': 'custom-date'}},
    )

    assert result == {'type': 'custom-date'}



from __future__ import annotations

import decimal
import typing

from django_filters import (
    Filter,
    BooleanFilter,
    ChoiceFilter,
    DateFilter,
    DateFromToRangeFilter,
    DateTimeFilter,
    DateTimeFromToRangeFilter,
    NumberFilter,
    TimeFilter,
    TypedChoiceFilter,
    MultipleChoiceFilter,
    TypedMultipleChoiceFilter,
)


def _get_filter_schema_choice(
    filter_field: typing.Union[
        ChoiceFilter, MultipleChoiceFilter, TypedChoiceFilter, TypedMultipleChoiceFilter
    ]
) -> dict:
    choice_type = 'string'

    try:
        choices = filter_field.field.choices
    except KeyError:
        return {'type': choice_type}

    choice_keys = [c for c, _ in choices]
    for type_, repr_ in (
        (bool, 'boolean'),
        (int, 'integer'),
        ((int, float, decimal.Decimal), 'number'),
    ):
        if all(isinstance(choice, type_) for choice in choice_keys):  # type: ignore
            choice_type = repr_
            break

    return {'type': choice_type, 'enum': choice_keys}


FILTER_MAP: typing.Dict[
    typing.Type[Filter], typing.Union[dict, typing.Callable[[Filter], dict]]
] = {
    BooleanFilter: {'type': 'boolean'},
    ChoiceFilter: _get_filter_schema_choice,
    MultipleChoiceFilter: _get_filter_schema_choice,
    TypedChoiceFilter: _get_filter_schema_choice,
    TypedMultipleChoiceFilter: _get_filter_schema_choice,
    DateFilter: {'type': 'string', 'format': 'date'},
    DateFromToRangeFilter: {'type': 'string', 'format': 'date'},
    DateTimeFilter: {'type': 'string', 'format': 'date-time'},
    DateTimeFromToRangeFilter: {'type': 'string', 'format': 'date-time'},
    NumberFilter: {'type': 'number'},
    TimeFilter: {'type': 'string', 'format': 'time'},
}


def get_filter_schema(filter_field: Filter, filter_map: dict = None) -> dict:
    filter_map = filter_map or FILTER_MAP
    field_parents = type(filter_field).mro()

    schema: typing.Union[dict, typing.Callable] = {'type': 'string'}
    for field_parent in field_parents[:-1]:
        try:
            schema = filter_map[field_parent]
            break
        except KeyError:
            pass

    if callable(schema):
        return schema(filter_field)
    return schema



from __future__ import annotations

from tests.test_unit.test_schema.stubs import DefaultViewSet


def test_override_operation_id(get_create_view_func):
    create_view = get_create_view_func('test', DefaultViewSet, 'test')

    view = create_view('/test/', 'GET')
    view.schema_operation_id_map = {'list': 'custom_id'}
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['operationId'] == 'custom_id'



from __future__ import annotations

import pytest
from django.core.exceptions import ImproperlyConfigured
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.filters import OrderingFilter

from restdoctor.rest_framework.schema import RestDoctorSchema
from restdoctor.rest_framework.schema.generators import RefsSchemaGenerator30, RefsSchemaGenerator31
from tests.test_unit.test_schema.stubs import (
    DefaultFilterSet,
    FilterSetWithLabels,
    FilterSetWithNoLabels,
    ListViewSetWithoutRequestSerializer,
    ListViewSetWithRequestSerializer,
)


@pytest.mark.parametrize(
    ('viewset', 'generator_class', 'expected_parameters'),
    [
        (
            ListViewSetWithRequestSerializer,
            RefsSchemaGenerator30,
            [
                {
                    'in': 'query',
                    'name': 'filter_uuid_field',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'uuid'},
                },
                {
                    'in': 'query',
                    'name': 'filter_field',
                    'required': True,
                    'schema': {'type': 'string', 'nullable': True, 'description': 'Help text'},
                },
            ],
        ),
        (
            ListViewSetWithRequestSerializer,
            RefsSchemaGenerator31,
            [
                {
                    'in': 'query',
                    'name': 'filter_uuid_field',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'uuid'},
                },
                {
                    'in': 'query',
                    'name': 'filter_field',
                    'required': True,
                    'schema': {'type': ['string', 'null'], 'description': 'Help text'},
                },
            ],
        ),
        (ListViewSetWithoutRequestSerializer, RefsSchemaGenerator30, []),
        (ListViewSetWithoutRequestSerializer, RefsSchemaGenerator31, []),
    ],
)
def test_schema_for_list_viewset(
    get_create_view_func, viewset, generator_class, expected_parameters
):
    create_view = get_create_view_func('test', viewset, 'test', generator_class=generator_class)

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['parameters'] == expected_parameters


@pytest.mark.parametrize(
    ('viewset', 'generator_class', 'expected_parameters'),
    [
        (
            ListViewSetWithRequestSerializer,
            RefsSchemaGenerator30,
            [
                {
                    'in': 'query',
                    'name': 'filter_uuid_field',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'uuid'},
                },
                {
                    'in': 'query',
                    'name': 'filter_field',
                    'required': True,
                    'schema': {'type': 'string', 'nullable': True, 'description': 'Help text'},
                },
            ],
        ),
        (
            ListViewSetWithRequestSerializer,
            RefsSchemaGenerator31,
            [
                {
                    'in': 'query',
                    'name': 'filter_uuid_field',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'uuid'},
                },
                {
                    'in': 'query',
                    'name': 'filter_field',
                    'required': True,
                    'schema': {'type': ['string', 'null'], 'description': 'Help text'},
                },
            ],
        ),
        (ListViewSetWithoutRequestSerializer, RefsSchemaGenerator30, []),
        (ListViewSetWithoutRequestSerializer, RefsSchemaGenerator31, []),
    ],
)
def test_get_request_serializer_filter_parameters(
    get_create_view_func, viewset, generator_class, expected_parameters
):
    list_view = get_create_view_func('test', viewset, 'test', generator_class=generator_class)

    view = list_view('/test/', 'GET')
    parameters = view.schema.get_request_serializer_filter_parameters('/test/', 'GET')

    assert parameters == expected_parameters


@pytest.mark.parametrize(
    ('filter_backends', 'filterset_class', 'expected_parameters'),
    [
        (
            [DjangoFilterBackend, OrderingFilter],
            DefaultFilterSet,
            [
                {
                    'description': 'UUID',
                    'in': 'query',
                    'name': 'uuid',
                    'required': False,
                    'schema': {'type': 'string'},
                },
                {
                    'description': 'Another Event timestamp',
                    'in': 'query',
                    'name': 'timestamp',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'date-time'},
                },
                {
                    'description': 'Which field to use when ordering the results.',
                    'in': 'query',
                    'name': 'ordering',
                    'required': False,
                    'schema': {'type': 'string'},
                },
            ],
        ),
        (
            [DjangoFilterBackend],
            FilterSetWithNoLabels,
            [
                {
                    'description': 'UUID',
                    'in': 'query',
                    'name': 'uuid',
                    'required': False,
                    'schema': {'type': 'string'},
                },
                {
                    'description': 'Event timestamp',
                    'in': 'query',
                    'name': 'my_model__timestamp',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'date-time'},
                },
                {
                    'description': 'Event timestamp',
                    'in': 'query',
                    'name': 'my_model__timestamp__in',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'date-time'},
                },
                {
                    'description': 'my another one model',
                    'in': 'query',
                    'name': 'my_another_one_model__isnull',
                    'required': False,
                    'schema': {'type': 'boolean'},
                },
                {
                    'description': 'Another One Event timestamp',
                    'in': 'query',
                    'name': 'my_another_one_model__timestamp',
                    'required': False,
                    'schema': {'format': 'date-time', 'type': 'string'},
                },
                {
                    'description': 'Another Event timestamp',
                    'in': 'query',
                    'name': 'created_at_date',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'date'},
                },
                {
                    'description': 'created_after',
                    'in': 'query',
                    'name': 'created_after',
                    'required': False,
                    'schema': {'type': 'string', 'format': 'date'},
                },
            ],
        ),
        (
            [DjangoFilterBackend],
            FilterSetWithLabels,
            [
                {
                    'description': 'UUID',
                    'in': 'query',
                    'name': 'uuid',
                    'required': False,
                    'schema': {'type': 'string'},
                },
                {
                    'description': 'Event timestamp',
                    'in': 'query',
                    'name': 'my_model__timestamp',
                    'required': False,
                    'schema': {'format': 'date-time', 'type': 'string'},
                },
                {
                    'description': 'Another One Event timestamp',
                    'in': 'query',
                    'name': 'my_another_one_model__timestamp',
                    'required': False,
                    'schema': {'format': 'date-time', 'type': 'string'},
                },
                {
                    'description': 'Created At Timestamp Label',
                    'in': 'query',
                    'name': 'created_at_date',
                    'required': False,
                    'schema': {'format': 'date', 'type': 'string'},
                },
                {
                    'description': 'Custom Method Label',
                    'in': 'query',
                    'name': 'created_after',
                    'required': False,
                    'schema': {'format': 'date', 'type': 'string'},
                },
            ],
        ),
    ],
)
def test_schema_for_viewset_with_filter_backend(
    get_create_view_func,
    viewset_with_filter_backends_factory,
    filter_backends,
    filterset_class,
    expected_parameters,
):
    viewset = viewset_with_filter_backends_factory(filter_backends, filterset_class)
    create_view = get_create_view_func('test', viewset, 'test')

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['parameters'] == expected_parameters


def test_schema_for_viewset_with_filter_backend_strict_schema(
    settings, get_create_view_func, viewset_with_filter_backends_factory
):
    settings.API_STRICT_SCHEMA_VALIDATION = True
    viewset = viewset_with_filter_backends_factory([DjangoFilterBackend], FilterSetWithLabels)
    create_view = get_create_view_func('test', viewset, 'test')
    view = create_view('/test/', 'GET')

    view.schema.get_operation('/test/', 'GET')


def test_schema_for_viewset_with_filter_backend_strict_schema_raises(
    settings, get_create_view_func, viewset_with_filter_backends_factory
):
    settings.API_STRICT_SCHEMA_VALIDATION = True
    viewset = viewset_with_filter_backends_factory([DjangoFilterBackend], FilterSetWithNoLabels)
    create_view = get_create_view_func('test', viewset, 'test')
    view = create_view('/test/', 'GET')

    with pytest.raises(ImproperlyConfigured):
        view.schema.get_operation('/test/', 'GET')


@pytest.mark.parametrize('ignore,is_expected_allow', ((True, False), (False, True)))
def test_schema_for_viewset_allows_filters(ignore, is_expected_allow, settings, mocker):
    mocker.patch(
        'restdoctor.rest_framework.schema.openapi.RestDoctorSchema.view', return_value=None
    )
    mocker.patch('rest_framework.schemas.utils.is_list_view', return_value=False)
    mocker.patch('rest_framework.schemas.openapi.AutoSchema.allows_filters', return_value=True)
    settings.API_IGNORE_FILTER_PARAMS_FOR_DETAIL = ignore

    is_allow = RestDoctorSchema().allows_filters('path', 'method')

    assert is_allow is is_expected_allow



from __future__ import annotations

import pytest

from restdoctor.rest_framework.schema import RefsSchemaGenerator


@pytest.mark.parametrize(
    ('api_format', 'expected_api_format'),
    [
        (('full', 'compact', 'compact-verbose'), ['full', 'compact', 'compact-verbose']),
        (('full', 'compact:{12,32}'), ['full', 'compact:12', 'compact:32']),
        (('compact:{12}',), ['compact:12']),
        (('test:{1,2,3,4}',), ['test:1', 'test:2', 'test:3', 'test:4']),
        (('full', 'compact:12'), ['full', 'compact:12']),
    ],
)
def test_schema_with_tags_success_case(settings, api_format, expected_api_format):
    settings.API_FORMATS = api_format

    refs_schema = RefsSchemaGenerator()

    assert refs_schema.api_formats == expected_api_format



from __future__ import annotations

import re
import contextlib
from typing import Optional, Dict

from django.conf import settings
from django.core.exceptions import (
    ObjectDoesNotExist,
    ValidationError as DjangoValidationError,
    ImproperlyConfigured,
)
from django.db.models import Model, QuerySet
from django.http import Http404
from rest_framework.generics import GenericAPIView as BaseGenericAPIView, get_object_or_404

from restdoctor.rest_framework.mixins import NegotiatedMixin


class GenericAPIView(NegotiatedMixin, BaseGenericAPIView):
    lookup_fields: Optional[Dict[str, str]] = None

    def get_object(self) -> Model:
        if self.lookup_fields is None:
            return self._simple_get_object()

        self._check_lookup_configuration()
        queryset = self._get_queryset_for_object()

        obj = None
        filter_value = self.kwargs[self.lookup_url_kwarg]
        for lookup_field, lookup_regex in self.lookup_fields.items():
            if re.match(lookup_regex, filter_value):
                filter_kwargs = {lookup_field: filter_value}
                with contextlib.suppress(
                    AttributeError, TypeError, ValueError, DjangoValidationError, ObjectDoesNotExist
                ):
                    obj = queryset.get(**filter_kwargs)

        if obj is None:
            raise Http404

        self.check_object_permissions(self.request, obj)

        return obj

    def _simple_get_object(self) -> Model:
        # copy paste from `rest_framework.generics.GenericAPIView.get_object`, but other `queryset`
        queryset = self._get_queryset_for_object()  # <-changed

        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' % (self.__class__.__name__, lookup_url_kwarg)
        )

        filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
        obj = get_object_or_404(queryset, **filter_kwargs)

        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    def _check_lookup_configuration(self) -> None:
        if not self.lookup_url_kwarg:
            raise ImproperlyConfigured(
                (
                    f'View {self.__class__.__name__} improperly configured. '
                    f'Set the `.lookup_url_kwarg` with `.lookup_fields`'
                )
            )

        if self.lookup_url_kwarg not in self.kwargs:
            raise ImproperlyConfigured(
                (
                    f'Expected view {self.__class__.__name__} to be called with a URL keyword argument '
                    f'named "{self.lookup_url_kwarg}". Fix your URL conf, or set the `.lookup_field` '
                    f'attribute on the view correctly.'
                )
            )

    def _get_queryset_for_object(self) -> QuerySet:
        qs = self.get_queryset()
        if settings.API_IGNORE_FILTER_PARAMS_FOR_DETAIL:
            return qs
        return self.filter_queryset(qs)



from __future__ import annotations

import typing

from django.core.management.base import BaseCommand, CommandParser
from django.utils.module_loading import import_string
from rest_framework import renderers

from restdoctor.rest_framework.schema import RefsSchemaGenerator
from restdoctor.rest_framework.schema.custom_types import SchemaGenerator


class Command(BaseCommand):
    help = 'Generates configured API schema for project.'  # noqa: A003, VNE003

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('--title', dest='title', default='', type=str)
        parser.add_argument('--url', dest='url', default=None, type=str)
        parser.add_argument('--description', dest='description', default=None, type=str)
        parser.add_argument(
            '--format',
            dest='format',
            choices=['openapi', 'openapi-json'],
            default='openapi',
            type=str,
        )
        parser.add_argument('--urlconf', dest='urlconf', default=None, type=str)
        parser.add_argument('--accept', dest='accept', default=None, type=str)
        parser.add_argument('--generator_class', dest='generator_class', default=None, type=str)
        parser.add_argument('--file', dest='file', default=None, type=str)

    def handle(self, *args: typing.Any, **options: typing.Any) -> None:
        if options['generator_class']:
            generator_class = import_string(options['generator_class'])
        else:
            generator_class = self.get_generator_class()
        generator = generator_class(
            url=options['url'],
            title=options['title'],
            description=options['description'],
            urlconf=options['urlconf'],
            accept=options['accept'],
        )
        schema = generator.get_schema(request=None, public=True)
        renderer = self.get_renderer(options['format'])
        output = renderer.render(schema, renderer_context={})

        if options['file']:
            with open(options['file'], 'wb') as output:
                output.write(output)
        else:
            self.stdout.write(output.decode())

    def get_renderer(self, fmt: str) -> renderers.BaseRenderer:
        renderer_cls = {
            'openapi': renderers.OpenAPIRenderer,
            'openapi-json': renderers.JSONOpenAPIRenderer,
        }[fmt]
        return renderer_cls()

    def get_generator_class(self) -> typing.Type[SchemaGenerator]:
        return RefsSchemaGenerator



import pytest

from restdoctor.rest_framework.generics import GenericAPIView


@pytest.mark.parametrize(
    'ignore,is_get_queryset_called,is_filter_queryset_called',
    ((True, True, False), (False, True, True)),
)
def test_get_queryset_for_object(
    ignore, is_get_queryset_called, is_filter_queryset_called, mocker, settings
):
    settings.API_IGNORE_FILTER_PARAMS_FOR_DETAIL = ignore
    get_queryset = mocker.patch('rest_framework.generics.GenericAPIView.get_queryset')
    filter_queryset = mocker.patch('rest_framework.generics.GenericAPIView.filter_queryset')

    GenericAPIView()._get_queryset_for_object()

    assert get_queryset.called is is_get_queryset_called
    assert filter_queryset.called is is_filter_queryset_called


@pytest.mark.parametrize('lookup_fields', (None, {'test': 'test'}))
def test_get_object_use_get_queryset_for_object(lookup_fields, mocker):
    mocker.patch('restdoctor.rest_framework.generics.GenericAPIView._check_lookup_configuration')
    get_queryset_for_object = mocker.patch(
        'restdoctor.rest_framework.generics.GenericAPIView._get_queryset_for_object'
    )
    view = GenericAPIView()
    view.lookup_fields = lookup_fields

    try:
        view.get_object()
    except AttributeError:
        pass

    assert get_queryset_for_object.called



from __future__ import annotations

import typing

import pytest

from restdoctor.utils.typing_inspect import is_list_type, is_optional_type, is_union_type

T = typing.TypeVar('T')


@pytest.mark.parametrize(
    ('annotation', 'expected_result'),
    [
        (typing.Union[int], False),
        (typing.Union[int, int], False),
        (typing.Union, True),
        (typing.Union[T, int], True),
        (typing.Union['T', int], True),
    ],
)
def test_is_union_type(annotation, expected_result):
    result = is_union_type(annotation)

    assert result == expected_result


@pytest.mark.parametrize(
    ('annotation', 'expected_result'),
    [
        (type(None), True),
        (typing.Optional[str], True),
        (typing.Optional[T], True),
        (typing.Optional['T'], True),
        (typing.Union[None], True),
        (typing.Union[int, None], True),
        (typing.Union[T, None], True),
        (typing.Union['T', None], True),
        (typing.Union[int, str], False),
        (typing.Union[T, str], False),
        (typing.Union['T', str], False),
        (typing.Union[int, typing.Optional[str]], True),
    ],
)
def test_is_optional_type(annotation, expected_result):
    result = is_optional_type(annotation)

    assert result == expected_result


@pytest.mark.parametrize(
    ('annotation', 'expected_result'),
    [
        (int, False),
        (dict, False),
        (set, False),
        (list, True),
        (typing.List[int], True),
        (typing.List[T], True),
        (typing.List['T'], True),
        (typing.Optional[list], True),
        (typing.Optional[typing.List], True),
        (typing.Optional[typing.List[T]], True),
        (typing.Optional[set], False),
        (typing.Optional[typing.Set], False),
        (typing.Optional[typing.Set[T]], False),
    ],
)
def test_is_list_type(annotation, expected_result):
    result = is_list_type(annotation)

    assert result == expected_result



from __future__ import annotations

from tests.test_unit.test_schema.stubs import MultipleSiblingParametersView


def test_schema_parameters_default_priority(get_create_view_func, settings):
    create_view = get_create_view_func('test', MultipleSiblingParametersView, 'test')
    settings.API_SCHEMA_PRIORITIZE_SERIALIZER_PARAMETERS = False

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['parameters'][2]['description'] == 'filter some field'


def test_schema_parameters_serializer_priority(get_create_view_func, settings):
    create_view = get_create_view_func('test', MultipleSiblingParametersView, 'test')
    settings.API_SCHEMA_PRIORITIZE_SERIALIZER_PARAMETERS = True

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['parameters'][2]['schema']['description'] == 'serializer some field'



from __future__ import annotations

import typing
from urllib.parse import urljoin

from django.conf import settings
from rest_framework.generics import GenericAPIView
from rest_framework.request import Request
from semver import VersionInfo

from restdoctor.rest_framework.schema.custom_types import SchemaGenerator
from restdoctor.rest_framework.schema.openapi import RestDoctorSchema
from restdoctor.rest_framework.schema.refs_registry import LocalRefsRegistry
from restdoctor.utils.api_format import get_available_format
from restdoctor.utils.media_type import parse_accept

if typing.TYPE_CHECKING:
    from restdoctor.rest_framework.custom_types import Handler
    from restdoctor.rest_framework.schema.custom_types import OpenAPISchema


class RefsSchemaGenerator(SchemaGenerator):
    openapi_version = None

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self.local_refs_registry = LocalRefsRegistry()

        if not self.openapi_version:
            self.openapi_version = VersionInfo.parse(settings.API_DEFAULT_OPENAPI_VERSION)

        self._operation_ids: typing.Dict[str, typing.Tuple[str, str]] = {}

        self.api_default_version = settings.API_DEFAULT_VERSION
        self.api_default_format = settings.API_DEFAULT_FORMAT
        self.api_default_content_type = f'application/vnd.{settings.API_VENDOR_STRING.lower()}'

        accept = kwargs.pop('accept', None)
        if accept:
            api_params = parse_accept(accept, settings.API_VENDOR_STRING.lower())

            api_version = api_params and api_params.version
            urlconf = settings.API_VERSIONS.get(api_version)
            if urlconf is None:
                raise Exception(f'Can not determine URLCONF for Accept string {accept}')
            self.api_version = api_version
            self.api_resource = api_params.resource_discriminator
            self.api_formats = [api_params.format]
            self.include_default_schema = False
        else:
            self.api_version = self.api_default_version
            urlconf = kwargs.pop('urlconf', None)
            if urlconf:
                for api_version, api_urlconf in settings.API_VERSIONS.items():
                    if api_urlconf == urlconf:
                        self.api_version = api_version
                        break

            self.api_resource = None
            self.api_formats = get_available_format(settings.API_FORMATS)
            self.include_default_schema = True

        kwargs['urlconf'] = urlconf
        super().__init__(*args, **kwargs)

    def get_paths(self, request: Request = None) -> typing.Optional[OpenAPISchema]:
        result: OpenAPISchema = {}

        paths, view_endpoints = self._get_paths_and_endpoints(request)

        if not paths:
            return None

        for path, method, view in view_endpoints:
            if not self.has_view_permissions(path, method, view):
                continue
            operation = view.schema.get_operation(path, method)

            operation_id = operation['operationId']
            if (
                operation_id in self._operation_ids
                and (method, path) != self._operation_ids[operation_id]
            ):
                operation_str = f'{view.__class__.__name__} {method, path}'
                action_name = view.schema.get_action_name(path, method)
                raise Exception(
                    f'Operation ID {operation_id} for view {operation_str} duplicates '
                    f'existing id for {self._operation_ids[operation_id]}. Consider adding '
                    f"schema_operation_id_map = {{ '{action_name}': '<your_custom_operation_id>' }} "
                    'to view/viewset to resolve this issue'
                )

            self._operation_ids[operation_id] = method, path

            if path.startswith('/'):
                path = path[1:]
            path = urljoin(self.url or '/', path)

            result.setdefault(path, {})
            result[path][method.lower()] = operation

        return result

    def create_view(
        self, callback: Handler, method: str, request: Request = None
    ) -> GenericAPIView:
        view = super().create_view(callback, method, request)
        schema_class = getattr(view, 'schema_class', RestDoctorSchema)
        view.schema = schema_class(generator=self)
        return view

    def get_error_schema(
        self, description: str = 'Описание ошибки', detailed: bool = False
    ) -> OpenAPISchema:
        schema: OpenAPISchema = {
            'type': 'object',
            'properties': {'message': {'type': 'string', 'description': description}},
        }
        if detailed:
            schema['properties']['errors'] = {'type': 'array', 'items': {'type': 'object'}}
        return schema

    def get_schema(
        self, request: Request = None, public: bool = False
    ) -> typing.Optional[OpenAPISchema]:
        self._initialise_endpoints()
        self.local_refs_registry.put_local_ref(
            '#/components/schemas/ErrorResponseSchemaDetailed', self.get_error_schema(detailed=True)
        )
        self.local_refs_registry.put_local_ref(
            '#/components/schemas/ErrorResponseSchema', self.get_error_schema()
        )
        self.local_refs_registry.put_local_ref(
            '#/components/schemas/NotFoundResponseSchema',
            self.get_error_schema(description='Ресурс не найден.'),
        )

        paths = self.get_paths(None if public else request)
        if not paths:
            return None

        schema = {'openapi': str(self.openapi_version), 'info': self.get_info(), 'paths': paths}
        components = self.local_refs_registry.get_components()
        if components:
            schema['components'] = components

        return schema


class RefsSchemaGenerator30(RefsSchemaGenerator):
    openapi_version = VersionInfo.parse('3.0.2')


class RefsSchemaGenerator31(RefsSchemaGenerator):
    openapi_version = VersionInfo.parse('3.1.0')
